import math
import matplotlib.pyplot as plt
from os import fspath
import subprocess
import tempfile
import seaborn as sns

from .config import config_context, Paper


def figure_setup(papernum=0):
    """Set all the sizes to the correct values.
    TODO: use tex fonts for all texts, if installed
    """
    cfg = config_context().papers[papernum]
    figure_setup = cfg.fig_settings
    # This sets reasonable defaults for font size in a paper
    # sns.set_context("paper")
    # Set the font to be serif, rather than sans
    # sns.set(font="serif")
    # Make the background white, and specify the specific font family
    rcparams = {
        # "font.serif": ["Times", "Palatino", "serif"],
        "figure.dpi": figure_setup.dpi,
        "font.size": figure_setup.font_size,
        "figure.titlesize": figure_setup.font_size,
        "axes.titlesize": figure_setup.font_size,
        "legend.fontsize": figure_setup.font_size,
        "legend.fontsize": figure_setup.label_size,
        "axes.labelsize": figure_setup.label_size,
        "xtick.labelsize": figure_setup.ticks_size,
        "ytick.labelsize": figure_setup.ticks_size,
        "axes.linewidth": figure_setup.axis_lw,
    }

    sns.set_theme(
        context=cfg.pubtype,
        style="white",
        font=figure_setup.font_family,
        rc=rcparams,
    )


def figsize(fig_width=None, fig_height=None, papernum=0, aspect="phi"):
    """helper to define figsize based on columns
    If no height is given, it is computed using the golden ratio.
    """
    paper_config = config_context().papers[papernum]
    assert (aspect in ["phi", "square"]) or isinstance(aspect, float)

    ratios = {
        "phi": (1.0 + math.sqrt(5.0)) / 2.0,
        "square": 1.0,
    }
    if aspect in ["phi", "square"]:
        aspect = 1.0 / ratios[aspect] if paper_config.columns == 1 else ratios[aspect]
    elif isinstance(aspect, float):
        pass
    else:
        raise TypeError("`aspect` must be one of {'phi', 'square'} or `float`")

    if fig_width is None:
        if isinstance(paper_config.pubtype, Paper) & paper_config.columns == 2:
            fig_width = paper_config.column_width
        else:
            fig_width = paper_config.width
    if fig_width is "full":
        fig_width = paper_config.width
    if fig_width is "column":
        fig_width = paper_config.column_width
    if fig_height is None:
        fig_height = fig_width / aspect

    return fig_width, fig_height


def save_fig(fig, file_name, fmt=None, tight=True, papernum=0):
    """Save a Matplotlib figure as EPS/PNG/PDF to the given path and trim it."""
    cfg = config_context().papers[papernum]
    fig_cfg = cfg.fig_settings
    if not fmt:
        fmt = file_name.suffix  # pathlib

    if fmt not in [".eps", ".png", ".pdf"]:
        raise ValueError("unsupported format: %s" % (fmt,))

    extension = fmt
    if not file_name.suffix == extension:
        file_name = file_name.with_suffix(extension)

    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_name = tmp_file.name + extension

    # save figure
    if tight:
        fig.savefig(tmp_name, dpi=fig_cfg.dpi, bbox_inches="tight")
    else:
        fig.savefig(tmp_name, dpi=fig_cfg.dpi)

    # trim it
    if fmt == ".eps":
        subprocess.call(
            "epstool --bbox --copy %s %s" % (tmp_name, fspath(file_name)), shell=True
        )
    elif fmt == ".png":
        subprocess.call(
            "convert {} -trim {}".format(tmp_name, fspath(file_name)), shell=True
        )
    elif fmt == ".pdf":
        subprocess.call("pdfcrop {} {}".format(tmp_name, fspath(file_name)), shell=True)


def args_to_str(args, skip=None):
    """
    transform an argparse object into a string for the purposes of naming in a
    reproducible way. Can be given a list of 'skipped' terms, which will be
    ignored (e.g. whether to save files is irrelevant if they're being saved).

    """
    skip = [] if skip is None else skip

    def abrev_list(s):
        if isinstance(s, list):
            return "-".join(map(str, s))
        else:
            return str(s)

    argstr = "_".join(
        [
            "".join([str(arg), abrev_list(getattr(args, arg))])
            for arg in vars(args)
            if str(arg) not in skip
        ]
    )
    return argstr


def plot_or_save(dir, args, fname, fig=None):
    """
    show plots live, or save plots to file
     Used in in an experiment run where `args.save` is present. Automatically
     names files using `args_to_str`.
    REQUIRES `logging`be defined!
    """
    import logging

    if fig is None:
        fig = plt.gcf()
    if args.save:
        argstr = args_to_str(args)
        fpath = (dir / "_".join([fname, argstr])).with_suffix(".pdf")
        save_fig(fig, fpath)
        logging.info(f"saving file: {fpath}")
    else:
        plt.show(block=False)
        logging.info(f"displaying plot: {fname}")



from typing import Dict, Union, Optional

from srsly import read_yaml

# from pydantic.dataclasses import dataclass
from pydantic import (
    BaseModel,
    BaseSettings,
    DirectoryPath,
    AnyHttpUrl,
    PositiveFloat,
    validator,
    constr,
)
from pathlib import Path
from enum import Enum, IntEnum
from pint import UnitRegistry

ureg = UnitRegistry()
Q = ureg.Quantity
Q_str = constr(regex="^[0-9]+.?[0-9]*\\s?[a-zA-Z_]+")
AcceptableDist = Union[PositiveFloat, Q_str, Q]


class PubType(str, Enum):
    paper = "paper"
    talk = "talk"
    poster = "poster"


class PaperColumns(IntEnum):
    single = 1
    double = 2


class FigureSettings(BaseModel):
    dpi: int = 200
    label_size: int = 8
    font_size: int = 8
    ticks_size: int = 6
    axis_lw: PositiveFloat = 0.6
    plot_lw: PositiveFloat = 1.5
    font_family: str = "serif"


def norm_distance_to_in(d: AcceptableDist) -> PositiveFloat:
    def to_quantity_inch(d: AcceptableDist) -> Q:
        if isinstance(d, str):
            d = Q(d)
        else:
            d = Q(d, "inch")
        return d

    q = to_quantity_inch(d)
    assert q.magnitude > 0

    return q.to("in").magnitude


class Publication(BaseSettings):
    pubtype: PubType
    folder: DirectoryPath
    height: PositiveFloat
    width: PositiveFloat
    git_url: Optional[AnyHttpUrl]
    fig_settings: Optional[FigureSettings]

    _norm_dim_to_inch = validator(
        "height", "width", pre=True, always=True, allow_reuse=True
    )(norm_distance_to_in)


class Paper(Publication):
    pubtype: PubType = PubType.paper
    columns: PaperColumns = PaperColumns.single
    column_width: PositiveFloat = None

    @validator("pubtype", pre=True, always=True)
    def coerce_pubtype(cls, typ):
        return PubType.paper

    @validator("column_width", pre=True, always=True)
    def col_leq_width(cls, colw, values):
        if colw == None and values["columns"] == 1:
            colw = values["width"]
        elif colw == None:
            raise TypeError("Must pass column width for 2-column papers!")
        colw = norm_distance_to_in(colw)
        assert colw <= values["width"], "column width must be strictly less than width"
        return colw


class Publications(BaseModel):
    __root__: Dict[str, Union[Paper, Publication]]
    # TODO: Discover w/ plugin system
    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        if isinstance(item, int):
            item = list(self.__root__.keys())[item]
        return self.__root__[item]


class DDRPRegistry(BaseModel):
    papers: Publications
    fig_settings: Optional[FigureSettings] = FigureSettings()


def config_context() -> DDRPRegistry:

    yml_loc = Path("ddrp.yml")
    config_dict = read_yaml(yml_loc)
    return DDRPRegistry.parse_obj(config_dict)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ddrp']

package_data = \
{'': ['*']}

install_requires = \
['Pint>=0.17,<0.18',
 'cookiecutter>=1.7.2,<2.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'srsly>=2.4.0,<3.0.0']

setup_kwargs = {
    'name': 'ddr-p',
    'version': '0.2.1',
    'description': 'data-driven research papers, making use of Python and LaTeX for automation and reproducibility.',
    'long_description': None,
    'author': 'Thurston Sexton',
    'author_email': 'thurston.sexton@nist.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
